#include "instance.h"
#include "tomax.h"
#include "misc_ray.h"

using namespace VR;

VRayMeshGeomInstance::VRayMeshGeomInstance(VRayMeshGeom *m, INode *node, VRayCore *vray, int renderID, int moving) {
	VRenderInstance::init(m, node, vray, renderID);

	tm=NULL;
	itm=NULL;
	radius=NULL;

	isMoving=moving;
}

void VRayMeshGeomInstance::freeMem(void) {
	delete[] tm;
	tm=NULL;

	delete[] itm;
	itm=NULL;

	delete[] radius;
	radius=NULL;

	numTMs=0;
}

VRayMeshGeomInstance::~VRayMeshGeomInstance(void) {
	freeMem();
}

void VRayMeshGeomInstance::renderBegin(TimeValue t, VRayCore *vray) {
	VRenderInstance::renderBegin(t, vray);
}

void VRayMeshGeomInstance::renderEnd(VRayCore *vray) {
	VRenderInstance::renderEnd(vray);
}

void VRayMeshGeomInstance::frameBegin(TimeValue t, VRayCore *vray) {
	VRenderInstance::frameBegin(t, vray);

	mesh=&dummyMesh;

	objToWorld=node->GetObjTMAfterWSM(t);
	Matrix3 worldToObj=Inverse(objToWorld);

	const VRaySequenceData &sdata=vray->getSequenceData();
	const VRayFrameData &fdata=vray->getFrameData();

	camToObj=toMatrix3(toTransform(worldToObj)*Transform(fdata.camToWorld));
	objToCam=toMatrix3(Transform(fdata.worldToCam)*toTransform(objToWorld));

	// Create the transformation matrices
	freeMem();
	numTMs=sdata.params.moblur.on? 2 : 1; // getNumGeometrySamples(vray);

	tm=new TraceTransform[numTMs];
	itm=new TraceTransform[numTMs];
	radius=new float[numTMs];

	VRayMeshGeom *maxObj=((VRayMeshGeom*) renderObject);

	if (numTMs==1) {
		tm[0]=toTransform(objToWorld);
		radius[0]=maxObj->pblock2->GetFloat(pb_radius, t);
	} else {
		for (int i=0; i<numTMs; i++) {
			TimeValue t=(TimeValue) (fdata.frameStart+(fdata.frameEnd-fdata.frameStart)*(float)i/(float)(numTMs-1));
			tm[i]=toTransform(node->GetObjTMAfterWSM(t));
			radius[i]=maxObj->pblock2->GetFloat(pb_radius, t);
		}
	}

	for (int i=0; i<numTMs; i++) itm[i]=inverse(tm[i]);
}

void VRayMeshGeomInstance::frameEnd(VRayCore *vray) {
	VRenderInstance::frameEnd(vray);
}

// From Shadeable
void VRayMeshGeomInstance::shade(VRayContext &vri) {
	VRenderInstance::fullShade((VRayInterface&) vri);
}

// From DefaultShadeData
int VRayMeshGeomInstance::getTexCoords(VR::Vector t[3], int channel, const VR::VRayContext &rc) {
	VRayMeshGeom *mobj=static_cast<VRayMeshGeom*>(renderObject);
	VR::MeshInterface *mface=mobj->getMeshInterface();
	VR::MeshVoxel *voxel=mface->getVoxel(rc.rayresult.extra0);

	VR::MeshChannel *texChan=voxel->getChannel(VERT_TEX_CHANNEL0+channel);
	if (!texChan) {
		mface->releaseVoxel(voxel);
		return false;
	}

	VR::VertGeomData *texCoords=(Vlado::VertGeomData*) texChan->data;
	VR::FaceTopoData *faces=(Vlado::FaceTopoData*) voxel->getChannel(texChan->depChannelID)->data;

	if (!texCoords || !faces) {
		t[0]=t[1]=t[2]=VUtils::Vector(0.0f, 0.0f, 0.0f);
		mface->releaseVoxel(voxel);
		return false;
	}

	VR::FaceTopoData &face=faces[rc.rayresult.faceIndex];

	t[0]=texCoords[face.v[0]];
	t[1]=texCoords[face.v[1]];
	t[2]=texCoords[face.v[2]];

	mface->releaseVoxel(voxel);
	return true;
}

int VRayMeshGeomInstance::getMtlID(const VR::VRayContext &rc) {
	const VR::RayResult &ri=rc.rayresult;

	VRayMeshGeom *mobj=static_cast<VRayMeshGeom*>(renderObject);
	VR::MeshInterface *mface=mobj->getMeshInterface();
	VR::MeshVoxel *voxel=mface->getVoxel(rc.rayresult.extra0);

	VR::MeshChannel *faceInfoChan=voxel->getChannel(FACE_INFO_CHANNEL);
	int res;
	if (faceInfoChan) {
		VR::FaceInfoData *faces=(VR::FaceInfoData*) faceInfoChan->data;
		res=faces[ri.faceIndex].mtlID;
	} else {
		res=0;
	}

	mface->releaseVoxel(voxel);
	return res;
}

void VRayMeshGeomInstance::getSmoothUVWbases(const VRayContext &vri, int channel, VR::Vector bases[3]) {
	getUVWbases(vri, channel, bases);
}

// From ShadeInstance
Vector VRayMeshGeomInstance::getBasePt(const VR::VRayContext &vri, BaseTime baseTime) { return vri.rayresult.wpoint; }
Vector VRayMeshGeomInstance::getBasePtObj(const VR::VRayContext &vri) { return itm[0]*vri.rayresult.wpoint; }

Vector VRayMeshGeomInstance::worldToObjectVec(const VR::VRayContext &vri, const Vector &d) { return itm[0].m*d; }
Vector VRayMeshGeomInstance::worldToObjectPt(const VR::VRayContext &vri, const Vector &p) { return itm[0]*p; }
Vector VRayMeshGeomInstance::objectToWorldVec(const VR::VRayContext &vri, const Vector &d) { return tm[0].m*d; }
Vector VRayMeshGeomInstance::objectToWorldPt(const VR::VRayContext &vri, const Vector &p) { return tm[0]*p; }

Vector VRayMeshGeomInstance::getBaseGNormal(const VR::VRayContext &vri, BaseTime baseTime) { return vri.rayresult.gnormal; }
Vector VRayMeshGeomInstance::getBaseNormal(const VR::VRayContext &vri, BaseTime baseTime) { return vri.rayresult.origNormal; }
Vector VRayMeshGeomInstance::getBaseBumpNormal(const VR::VRayContext &vri, BaseTime baseTime) { return vri.rayresult.normal; }

TracePoint VRayMeshGeomInstance::getShadowPt(const VR::VRayContext &vri) { return vri.rayresult.wpoint; }
Vector VRayMeshGeomInstance::getVelocity(const VR::VRayContext &vri) { return Vector(0,0,0); }

void VRayMeshGeomInstance::compileGeometry(VR::VRayCore *vray) {
	VRayMeshGeom *mobj=(VRayMeshGeom*) renderObject;

	VR::RayServerInterface *rayserver=vray->getRayServer();
	if (!isMoving) {
		staticPrimitive.init(mobj->getMeshInterface(), tm[0], vray->getSequenceData().threadManager, static_cast<VUtils::GeometryGenerator*>(this));
		rayserver->storeStaticPrimitive(&staticPrimitive, NULL, rayserver->getNewRenderID());
	} else {
		movingPrimitive.init(mobj->getMeshInterface(), numTMs, tm, vray->getSequenceData().threadManager, static_cast<VUtils::GeometryGenerator*>(this), vray->getSequenceData().params.moblur.duration);
		rayserver->storeMovingPrimitive(&movingPrimitive, NULL, rayserver->getNewRenderID());
	}
}

void VRayMeshGeomInstance::setIntersectionData(RSRay &rsray, void *isd) {
	VR::IntersectionData &isData=*((VR::IntersectionData*) isd);
	VR::RSIntersection &is=rsray.is;

	int faceIndex=rsray.is.extrai[0];

	isData.primitive=is.primitive;
	isData.skipTag=is.skipTag;
	isData.faceIndex=faceIndex;
	isData.sb=(VR::Shadeable*) this;
	isData.sd=(VR::ShadeData*) this;
	isData.si=(VR::ShadeInstance*) this;
	isData.volume=NULL;

	isData.bary=is.bary;
	isData.wpointCoeff=is.t;
	isData.gnormal=(static_cast<VR::GeometryGenerator*>(rsray.is.primitive->owner))->getGNormal(rsray);
	isData.normal=(static_cast<VR::GeometryGenerator*>(rsray.is.primitive->owner))->getNormal(rsray);
	isData.wpoint=rsray.p+VUtils::TracePoint(rsray.dir)*rsray.is.t;

	VR::Vector p[3];
	VR::MeshVoxel *voxel=NULL;
	if (!isMoving) {
		VR::StaticMeshVoxelPrimitive *mesh=static_cast<VR::StaticMeshVoxelPrimitive*>(is.primitive->owner);
		voxel=mesh->voxel;
	
		VR::FaceTopoData *faces=voxel->getFaceTopoData();
		VR::VertGeomData *verts=voxel->getVertGeomData();
		const VR::Transform &tm=((VR::StaticMeshInterfacePrimitive*) mesh->owner)->tm;

		p[0]=tm*(verts[faces[faceIndex].v[0]]);
		p[1]=tm*(verts[faces[faceIndex].v[1]]);
		p[2]=tm*(verts[faces[faceIndex].v[2]]);
	} else {
		VR::MovingMeshVoxelPrimitive *mesh=static_cast<VR::MovingMeshVoxelPrimitive*>(is.primitive->owner);
		VR::MovingMeshInterfacePrimitive *meshOwner=static_cast<VR::MovingMeshInterfacePrimitive*>(mesh->owner);
		voxel=mesh->voxel;
	
		VR::FaceTopoData *faces=voxel->getFaceTopoData();
		VR::VertGeomData *verts=voxel->getVertGeomData();

		const VR::Transform &tm0=meshOwner->tms[mesh->timeIndex];
		const VR::Transform &tm1=meshOwner->tms[mesh->timeIndex+1];
		float k=(rsray.time-meshOwner->times[mesh->timeIndex])/(meshOwner->times[mesh->timeIndex+1]-meshOwner->times[mesh->timeIndex]);

		for (int i=0; i<3; i++) {
			VR::Vector p0=tm0*verts[faces[faceIndex].v[i]];
			VR::Vector p1=tm1*verts[faces[faceIndex].v[i]];
			p[i]=p0*(1.0f-k)+p1*k;
		}
	}

	isData.faceBase=p[0];
	isData.faceEdge0=p[1]-p[0];
	isData.faceEdge1=p[2]-p[0];
	isData.extra0=voxel->index; // This is the index of the voxel 

	isData.surfaceProps=&surfaceProps;
}
