#ifndef __INSTANCE_H__
#define __INSTANCE_H__

#include "vraymeshgeom.h"
#include "meshprimitives.h"

/*
	VRaySphereInstance: the actual instance that generates geometry.
	This derives StaticPrimitive which has the actual intersectable sphere as
	a sub-primitive. This is because all intersectable static primitives must
	have an expandable parent - the setIntersectionData() method of this parent
	will be called to fill in the intersection structure.

	This is different from infinite primitives, which should not have a parent
	but must be directly intersectable.

	The actual primitive that does the intersection is StaticSpherePrim.
*/

#if VRAY_DLL_VERSION < 0x20001
namespace VUtils {
typedef Transform TraceTransform;
}
#endif

class VRayMeshGeomInstance:
	public VR::VRenderInstance,
	public VR::Shadeable,
	public VR::DefaultShadeData,
	public VR::ShadeInstance,
	public VR::GeometryGenerator
{
protected:
	// Transformation of the sphere
	int numTMs;
	VR::TraceTransform *tm, *itm;
	float *radius;

	Mesh dummyMesh; // Dummy 3dsmax mesh

	int firstPrimID;

	VRayMeshGeom *meshGeom;

	void freeMem(void);

	int isMoving; // true if this is a motion-blurred instance
	VR::StaticMeshInterfacePrimitive staticPrimitive;
	VR::MovingMeshInterfacePrimitive movingPrimitive;
public:
	VRayMeshGeomInstance(VRayMeshGeom *m, INode *node, VR::VRayCore *vray, int renderID, int isMoving);
	virtual ~VRayMeshGeomInstance(void);

	// From VRenderInstance
	void renderBegin(TimeValue t, VR::VRayCore *vray);
	void frameBegin(TimeValue t, VR::VRayCore *vray);
	void frameEnd(VR::VRayCore *vray);
	void renderEnd(VR::VRayCore *vray);

	// From RenderInstance
	Interval MeshValidity(void) { return FOREVER; }
	Point3 GetFaceNormal(int faceNum) { return Point3(0,0,1); }
	Point3 GetFaceVertNormal(int faceNum, int vertNum) { return Point3(0,0,1); }
	void GetFaceVertNormals(int faceNum, Point3 n[3]) {}
	Point3 GetCamVert(int vertNum) { return Point3(0,0,0); }
	void GetObjVerts(int fnum, Point3 obp[3]) {}
	void GetCamVerts(int fnum, Point3 cp[3]) {}

	// From Shadeable
	void shade(VR::VRayContext &vri);

	// From DefaultShadeData
	int getTexCoords(VR::Vector t[3], int channel, const VR::VRayContext &rc);
	int getMtlID(const VR::VRayContext &vri);
	int getGBufID(void) { return 0; }
	int getSurfaceRenderID(const VR::VRayContext &rc) { return VRenderInstance::renderID; }
	int getSurfaceHandle(const VR::VRayContext &rc) { return VRenderInstance::nodeHandle; }
	int getMaterialRenderID(const VR::VRayContext &rc) { return 0; }
	void getSmoothUVWbases(const VR::VRayContext &vri, int channel, VR::Vector bases[3]);

	// From ShadeInstance
	VR::Vector getBasePt(const VR::VRayContext &vri, BaseTime baseTime);
	VR::Vector getBasePtObj(const VR::VRayContext &vri);

	VR::Vector getBaseGNormal(const VR::VRayContext &vri, BaseTime baseTime);
	VR::Vector getBaseNormal(const VR::VRayContext &vri, BaseTime baseTime);
	VR::Vector getBaseBumpNormal(const VR::VRayContext &vri, BaseTime baseTime);

	VR::Vector worldToObjectVec(const VR::VRayContext &vri, const VR::Vector &pt);
	VR::Vector worldToObjectPt(const VR::VRayContext &vri, const VR::Vector &pt);

	VR::Vector objectToWorldVec(const VR::VRayContext &vri, const VR::Vector &vec);
	VR::Vector objectToWorldPt(const VR::VRayContext &vri, const VR::Vector &pt);
	VR::TracePoint getShadowPt(const VR::VRayContext &vri);
	VR::Vector getVelocity(const VR::VRayContext &vri);

	// From VRenderInstance
	void compileGeometry(VR::VRayCore *vray);

	// From GeometryGenerator
	void setIntersectionData(VR::RSRay &rsray, void *isData);
};

#endif